/*
    (C) 1995-2001 AROS - The Amiga Research OS
    $Id: avail.c,v 1.2 2007/09/09 09:14:41 piru Exp $
 
    Desc: Avail CLI command
    Lang: English
*/

/******************************************************************************
 
 
    NAME
 
        Avail [CHIP | FAST | TOTAL | FLUSH]
 
    SYNOPSIS
 
        CHIP/S, FAST/S, TOTAL/S, FLUSH/S
 
    LOCATION
 
        Workbench:C
 
    FUNCTION
 
        Give a summary of the memory usage and availability in the system.
	To free up unused memory that still may be allocated (libraries,
	devices, fonts and such present in memory but whcih are currently
	not in use), use the FLUSH option.
 
    INPUTS
 
        CHIP   --  show only "chip" memory
	FAST   --  show only "fast" memory
	TOTAL  --  show information on memory regardless of type
	FLUSH  --  remove unnecessary things residing in memory
 
    RESULT
 
    NOTES
 
        "Chip" and "fast" memory are associated with the Amiga computer
	and may not be applicable on your hardware platform.
 
    EXAMPLE
 
    BUGS
 
    SEE ALSO
 
    INTERNALS
 
    HISTORY
 
******************************************************************************/

#include <exec/types.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#include <dos/dos.h>
#include <dos/dosextens.h>

#include <proto/exec.h>
#include <proto/dos.h>

#include "avail_version.h"

static const char version[] = VERSTAG;


#define  ARG_TEMPLATE  "CHIP/S,FAST/S,TOTAL/S,FLUSH/S,H=HUMAN/S"

enum
{
    ARG_CHIP = 0,
    ARG_FAST,
    ARG_TOTAL,
    ARG_FLUSH,
    ARG_HUMAN,
    NOOFARGS
};


struct localdata
{
    struct ExecBase	*SysBase;
    struct DosLibrary	*DOSBase;
    BOOL  aChip;
    BOOL  aFast;
    BOOL  aTotal;
    BOOL  aFlush;
    BOOL  aHuman;
};

#define SysBase ld->SysBase
#define DOSBase ld->DOSBase
#define aChip   ld->aChip
#define aFast   ld->aFast
#define aTotal  ld->aTotal
#define aFlush  ld->aFlush
#define aHuman  ld->aHuman


LONG printm(struct localdata *ld, CONST_STRPTR head, ULONG *array, LONG num);

int main(void)
{
    struct localdata _ld, *ld = &_ld;
    LONG           error = RETURN_FAIL;

    SysBase = *((struct ExecBase **) 4);
    DOSBase = (struct DosLibrary *) OpenLibrary("dos.library", 37);
    if (DOSBase)
    {
        BOOL bPrintErr = TRUE;
        IPTR           args[NOOFARGS] = { (IPTR)FALSE,
                                          (IPTR)FALSE,
                                          (IPTR)FALSE,
                                          (IPTR)FALSE,
                                          (IPTR)FALSE };
        struct RDArgs *rda;

        rda = ReadArgs(ARG_TEMPLATE, args, NULL);

        if (rda)
        {
            ULONG chip[4], fast[4], total[4];

            aChip  = args[ARG_CHIP] ? TRUE : FALSE;
            aFast  = args[ARG_FAST] ? TRUE : FALSE;
            aTotal = args[ARG_TOTAL] ? TRUE : FALSE;
            aFlush = args[ARG_FLUSH] ? TRUE : FALSE;
            aHuman = args[ARG_HUMAN] ? TRUE : FALSE;

            if (aChip + aFast + aTotal > 1)
            {
                PutStr("only one of CHIP, FAST, or TOTAL allowed\n");
                bPrintErr = FALSE;
                error = RETURN_WARN;
            }
            else
            {
                LONG res;

                error = RETURN_OK;

                if (aFlush)
                {
                    APTR Mem;

                    Mem = AllocVec(0x7ffffff0, MEMF_PUBLIC);
                    if (Mem)
                        FreeVec(Mem);
                }

                if (aChip)
                {
                    chip[0] = AvailMem(MEMF_CHIP);

                    res = printm(ld, NULL, chip, 1);
                    if (res < 0)
                    {
                        error = RETURN_ERROR;
                    }
                }
                else if (aFast)
                {
                    fast[0] = AvailMem(MEMF_FAST);

                    res = printm(ld, NULL, fast, 1);
                    if (res < 0)
                    {
                        error = RETURN_ERROR;
                    }
                }
                else if (aTotal)
                {
                    total[0] = AvailMem(MEMF_ANY);

                    res = printm(ld, NULL, total, 1);
                    if (res < 0)
                    {
                        error = RETURN_ERROR;
                    }
                }
                else
                {
                    Forbid();

                    chip[0]  = AvailMem(MEMF_CHIP);
                    chip[2]  = AvailMem(MEMF_CHIP | MEMF_TOTAL);
                    chip[3]  = AvailMem(MEMF_CHIP | MEMF_LARGEST);
                    fast[0]  = AvailMem(MEMF_FAST);
                    fast[2]  = AvailMem(MEMF_FAST | MEMF_TOTAL);
                    fast[3]  = AvailMem(MEMF_FAST | MEMF_LARGEST);
                    total[0] = AvailMem(MEMF_ANY);
                    total[2] = AvailMem(MEMF_ANY | MEMF_TOTAL);
                    total[3] = AvailMem(MEMF_ANY | MEMF_LARGEST);

                    Permit();

#ifdef __MORPHOS__
                    /* If Pegasos & fake chipmem header, adjust values - Piru */
                    if (SysBase->MaxLocMem == 0 && chip[2])
                    {
                        /* This is a bit sucky, relying on the internal
                           AvailMem implementation... - Piru */
                        total[2] -= chip[2] + sizeof(struct MemHeader);
                        chip[0] = chip[2] = chip[3]= 0;
                    }
#endif
                    chip[1]  = chip[2] - chip[0];
                    fast[1]  = fast[2] - fast[0];
                    total[1] = total[2] - total[0];

#if 1
                    if (PutStr("Type   Available    In-Use   Maximum   Largest\n") < 0 ||
                        (SysBase->MaxLocMem == 0 ? 0 : printm(ld, "chip", chip, 4)) < 0 ||
                        printm(ld, "fast", fast, 4) < 0 ||
                        printm(ld, "total", total, 4) < 0)
                    {
                        error = RETURN_ERROR;
                    }
#else
                    if (PutStr("Type   Available    In-Use   Maximum   Largest\n") < 0 ||
                        (SysBase->MaxLocMem == 0 ? 0 : VPrintf("chip%11lu %9lu %9lu %9lu\n", chip)) < 0 ||
                        VPrintf("fast%12lu %9lu %9lu %9lu\n", fast) < 0 ||
                        VPrintf("total%11lu %9lu %9lu %9lu\n", total) < 0)
                    {
                        error = RETURN_ERROR;
                    }
#endif
                }
            }

            FreeArgs(rda);
        }
        else
        {
            error = RETURN_FAIL;
        }

        if (error != RETURN_OK && bPrintErr)
        {
            PrintFault(IoErr(), "Avail");
        }

        CloseLibrary((struct Library *) DOSBase);
    }

    return error;
}


static
void fmtlarge(struct localdata *ld, UBYTE *buf, ULONG num)
{
    UQUAD d;
    UBYTE ch;
    struct
    {
        ULONG val;
        LONG  dec;
    } array =
    {
        num,
        0
    };

    if (num >= 1073741824)
    {
        array.val = num >> 30;
        d = ((UQUAD)num * 10 + 536870912) / 1073741824;
        array.dec = d % 10;
        ch = 'G';
    }
    else if (num >= 1048576)
    {
        array.val = num >> 20;
        d = ((UQUAD)num * 10 + 524288) / 1048576;
        array.dec = d % 10;
        ch = 'M';
    }
    else if (num >= 1024)
    {
        array.val = num >> 10;
        d = (num * 10 + 512) / 1024;
        array.dec = d % 10;
        ch = 'K';
    }
    else
    {
        array.val = num;
        array.dec = 0;
        d = 0;
        ch = 'B';
    }

    if (!array.dec && (d > array.val * 10))
    {
        array.val++;
    }

#ifndef __MORPHOS__
#fail "Uses RawDoFmt NULL putch func"
#endif
    RawDoFmt(array.dec ? "%lu.%lu" : "%lu", &array, NULL, buf);
    while (*buf) { buf++; }
    *buf++ = ch;
    *buf   = '\0';
}

LONG printm(struct localdata *ld, CONST_STRPTR head, ULONG *array, LONG num)
{
    LONG res = -1;
    CONST_STRPTR fmt;
    UBYTE buf[10];

    if (head)
    {
        LONG len = 16 - strlen(head);
#ifndef __MORPHOS__
#fail "Uses RawDoFmt NULL putch func"
#endif
        RawDoFmt(aHuman ? "%%%lds" : "%%%ldlu", &len, NULL, buf);
        fmt = buf;
        PutStr(head);
    }
    else
    {
        fmt = aHuman ? "%9s" : "%9lu";
    }

    if (aHuman)
    {
        if (num == 1)
        {
            fmtlarge(ld, buf, *array);
            res = PutStr(buf);
        }
        else
        {
            while (num--)
            {
                UBYTE tmp[10];

                fmtlarge(ld, tmp, *array);
                res = Printf(fmt, (ULONG) tmp);
                if (res < 0)
                    break;

                fmt = " %9s";
                array++;
            }
        }
    }
    else
    {
        if (num == 1)
        {
            res = VPrintf("%lu", array);
        }
        else
        {
            while (num--)
            {
                res = VPrintf(fmt, array);
                if (res < 0)
                    break;

                fmt = " %9lu";
                array++;
            }
        }
    }
    PutStr("\n");

    return res;
}
