/*
    (C) 1997-2001 AROS - The Amiga Research OS
    $Id: date.c,v 1.5 2018/08/24 16:57:31 piru Exp $
 
    Desc: Date CLI command
    Lang: English
*/

#include <exec/devices.h>
#include <exec/errors.h>
#include <exec/io.h>
#include <exec/memory.h>
#include <exec/semaphores.h>
#include <dos/dos.h>
#include <dos/datetime.h>
#include <dos/exall.h>
#include <devices/timer.h>
#include <libraries/locale.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/locale.h>

#include <stdio.h>
#include <string.h>

#include "date_version.h"

static const char version[] = VERSTAG;

#define ARG_STRING "DAY,DATE,TIME,TO=VER/K,LFORMAT/K"
enum
{
    ARG_DAY,
    ARG_DATE,
    ARG_TIME,
    ARG_VER,
    ARG_LFORMAT,
    ARG_COUNT
};

#ifndef FORMAT_DEF
#define FORMAT_DEF 4
#endif


static int printdate(STRPTR *filename_lformat, struct DosLibrary *DOSBase, struct ExecBase *SysBase);
static int setdate(STRPTR *day_date_time, struct DosLibrary *DOSBase, struct ExecBase *SysBase);

int main(void)
{
    struct ExecBase    *SysBase;
    struct DosLibrary  *DOSBase;
    int                error = RETURN_OK;

    SysBase = *((struct ExecBase **) 4);
    DOSBase = (struct DosLibrary *) OpenLibrary("dos.library", 37);
    if (DOSBase)
    {
        STRPTR             args[ARG_COUNT] = {NULL};
        struct RDArgs      *rda;

        rda = ReadArgs(ARG_STRING, (IPTR *) args, NULL);

        if (rda)
        {
            if (args[ARG_DAY] || args[ARG_DATE] || args[ARG_TIME])
            {
                if ((error = setdate(args, DOSBase, SysBase) == RETURN_OK))
                {
                    if (args[ARG_VER])
                    {
                        printdate(&args[ARG_VER], DOSBase, SysBase);
                    }
                }
            }
            else
            {
                error = printdate(&args[ARG_VER], DOSBase, SysBase);
            }

            FreeArgs(rda);
        }
        else
        {
            PrintFault(IoErr(), "Date");
            error = RETURN_FAIL;
        }

        CloseLibrary((struct Library *) DOSBase);
    }

    return error;
}

static LONG chrcount(STRPTR s, UBYTE c)
{
    UBYTE sc;
    LONG retval = 0;

    while((sc = *s++))
    {
        if (sc == c)
        {
            retval++;
        }
    }

    return retval;
}

static int setdate(STRPTR *day_date_time, struct DosLibrary *DOSBase, struct ExecBase *SysBase)
{
    int                error = RETURN_OK;
    BYTE               timererror;
    struct timerequest *timerReq;
    struct MsgPort     *timerMP;
    struct DateTime    dt;
    WORD               i, count;
    UBYTE              fulltime[9];
    STRPTR             realtime = NULL, realdate = NULL;

    for (i = 0; i < 3; i++)
    {
        if (!day_date_time[i])
        {
            continue;
        }

        if (chrcount(day_date_time[i], '-'))
        {
            /* must be date */

            realdate = day_date_time[i];
        }
        else if ((count = chrcount(day_date_time[i], ':')))
        {
            /* must be time */
            if (count == 1)
            {
                /* seconds are missing */

                if (strlen(day_date_time[i]) <= 5)
                {
                    strcpy(fulltime, day_date_time[i]);
                    strcat(fulltime, ":00");
                    realtime = fulltime;
                }
                else
                {
                    realtime = day_date_time[i];
                }
            }
            else
            {
                realtime = day_date_time[i];
            }
        }
        else
        {
            /* must be week day name */

            if (!realdate)
            {
                realdate = day_date_time[i];
            }
        }
    }

    timerMP = CreateMsgPort();
    if (timerMP)
    {
        timerReq = (struct timerequest *)CreateIORequest(timerMP, sizeof(struct timerequest));

        if (timerReq)
        {
            timererror = OpenDevice(TIMERNAME, UNIT_VBLANK,
                                    &timerReq->tr_node, 0L);
            if (!timererror)
            {
                dt.dat_Format  = FORMAT_DOS;
                dt.dat_Flags   = DTF_FUTURE;
                dt.dat_StrDay  = NULL; /* StrToDate ignores this anyway */
                dt.dat_StrDate = realdate;
                dt.dat_StrTime = realtime;

                DateStamp(&dt.dat_Stamp);

                if ((!realdate && !realtime) || !StrToDate(&dt))
                {
                    PutStr("***Bad args:\n"
                           "- use DD-MMM-YY or <dayname> or yesterday etc. to set date\n"
                           "      HH:MM:SS or HH:MM to set time\n");

                    error = RETURN_FAIL;
                }
                else
                {
                    timerReq->tr_time.tv_secs = dt.dat_Stamp.ds_Days * 60 * 60 * 24 +
                                                dt.dat_Stamp.ds_Minute * 60 +
                                                dt.dat_Stamp.ds_Tick / TICKS_PER_SECOND;
                    timerReq->tr_time.tv_micro = 0;
                    timerReq->tr_node.io_Command = TR_SETSYSTIME;
                    timerReq->tr_node.io_Flags |= IOF_QUICK;

                    if (DoIO(&timerReq->tr_node) != 0)
                    {
                        PutStr("Date: Unable to set system time\n");
                        error = RETURN_FAIL;
                    }
                }

                CloseDevice(&timerReq->tr_node);
            }
            else
            {
                PutStr("Date: Error opening timer.device\n");
                error = RETURN_FAIL;
            }

            DeleteIORequest(&timerReq->tr_node);
        }
        else
        {
            PutStr("Date: Error creating timerequest\n");
            error = RETURN_FAIL;
        }

        DeleteMsgPort(timerMP);
    }
    else
    {
        PutStr("Date: Error creating MsgPort\n");
        error = RETURN_FAIL;
    }

    return error;
}


struct myhook
{
    struct Hook        hook;
    struct DosLibrary *DOSBase;
    BPTR               file;
};

static void putchfunc(struct myhook *Hook,
                      struct Locale *Loc,
                      UBYTE ch)
{
    if (ch)
    {
        struct DosLibrary *DOSBase = Hook->DOSBase;

        FPutC(Hook->file, ch);
    }
}

extern const struct EmulLibEntry HookEntry;

static int printdate(STRPTR *filename_lformat, struct DosLibrary *DOSBase, struct ExecBase *SysBase)
{
    BPTR file = Output();
    int ownfile = FALSE;
    int error = RETURN_OK;
    struct DateTime dt;

    if (filename_lformat[0])
    {
        file = Open(filename_lformat[0], MODE_NEWFILE);
        ownfile = TRUE;
    }

    if (file)
    {
        DateStamp(&dt.dat_Stamp);

        if (filename_lformat[1])
        {
             struct Library *LocaleBase;

             LocaleBase = OpenLibrary("locale.library", 38);
             if (LocaleBase)
             {
                 struct Locale *locale = OpenLocale(NULL);
                 struct myhook putchhook;

                 putchhook.hook.h_Entry    = (HOOKFUNC) &HookEntry;
                 putchhook.hook.h_SubEntry = (HOOKFUNC) putchfunc;
                 putchhook.DOSBase         = DOSBase;
                 putchhook.file            = file;

                 FormatDate(locale, filename_lformat[1],
                            &dt.dat_Stamp, &putchhook.hook);

                 putchfunc(&putchhook, NULL, '\n');

                 CloseLocale(locale);

                 CloseLibrary(LocaleBase);
             }
             else
             {
                 Printf("Date: Error opening locale.library\n");
                 error = RETURN_FAIL;
             }
        }
        else
        {
            UBYTE dowstring[LEN_DATSTRING * 2], datestring[LEN_DATSTRING * 2],
                  timestring[LEN_DATSTRING * 2], resstring[LEN_DATSTRING * 6 + 1];
            int pos = 0;

            dt.dat_Format  = FORMAT_DOS;
            dt.dat_Flags   = 0;
            dt.dat_StrDay  = dowstring;
            dt.dat_StrDate = datestring;
            dt.dat_StrTime = timestring;
            DateToStr(&dt);

            CopyMem(dowstring, resstring, strlen(dowstring));
            pos += strlen(dowstring);
            resstring[pos++] = ' ';
            CopyMem(datestring, resstring + pos, strlen(datestring));
            pos += strlen(datestring);
            resstring[pos++] = ' ';
            CopyMem(timestring, resstring + pos, strlen(timestring));
            pos += strlen(timestring);
            resstring[pos++] = '\n';

            if (Write(file, resstring, pos) < pos)
            {
                PrintFault(IoErr(), "Date");
                error = RETURN_FAIL;
            }
        }

        if (ownfile)
        {
            Close(file);
        }
    }
    else
    {
        PrintFault(IoErr(), "Date");
        error = RETURN_FAIL;
    }

    return error;
}

